proc xml2list xml {
    regsub {^<\?xml[^>]*\?>} $xml {} xml
    regsub -all {>\s*<} [string trim $xml " \n\t<>"] "\} \{" xml
    set xml [string map {> "\} \{#text \{" < "\}\} \{"}  $xml]

    set res ""   ;# string to collect the result   
    set stack {} ;# track open tags
    set rest {}
    foreach item "{$xml}" {
        switch -regexp -- $item {
            ^# {append res "{[lrange $item 0 end]} " ; #text item}
            ^/ {
                regexp {/(.+)} $item -> tagname ;# end tag
                set expected [lindex $stack end]
                if {$tagname!=$expected} {error "$item != $expected"}
                set stack [lrange $stack 0 end-1]
                append res "\}\} "
        }
            /$ { # singleton - start and end in one <> group
                regexp {([^ ]+)( (.+))?/$} $item -> tagname - rest
                set rest [lrange [string map {= " "} $rest] 0 end]
                append res "{$tagname [list $rest] {}} "
            }
            default {
                set tagname [lindex $item 0] ;# start tag
                set rest [lrange [string map {= " "} $item] 1 end]
                lappend stack $tagname
                append res "\{$tagname [list $rest] \{"
            }
        }
        if {[llength $rest]%2} {error "att's not paired: $rest"}
    }
    if [llength $stack] {error "unresolved: $stack"}
    string map {"\} \}" "\}\}"} [lindex $res 0]
}

proc list2xml list {
    switch -- [llength $list] {
        2 {lindex $list 1}
        3 {
            foreach {tag attributes children} $list break
            set res <$tag
            foreach {name value} $attributes {
                append res " $name=\"$value\""
            }
            if [llength $children] {
                append res >
                foreach child $children {
                    append res [list2xml $child]
                }
                append res </$tag>
            } else {append res />}
        }
        default {error "could not parse $list"}
    }
}

proc xml_get_node { l path } {
    if { [llength $path] >= 1 } {
        if { [llength $l] != 3 } {
            error "Expected list of 3 elements but got [llength $l]"
        }
        set child_nodes [lindex $l 2]
        set nextNodeName [lindex $path 0]
        foreach n $child_nodes {
            if { [llength $n] != 3 } {
                error "Expected list of 3 elements but got [llength $n]"
            }
            set nodeName [lindex $n 0]
            if { "$nodeName" == "$nextNodeName" } {
                return [xml_get_node $n [lrange $path 1 end]]
            }
        }
        error "Node not found: $nextNodeName"
    } else {
        return $l
    }
}

proc xml_get_attribute { l path } {
    if { [llength $path] > 1 } {
        set node [xml_get_node $l [lrange $path 0 end-1]]
    } else {
        set node $l
    }
    set attrName [lindex $path end]
    array set attrs [lindex $node 1]
    if { ![info exist attrs($attrName)] } {
        error "Attribute \"$attrName\" not found on node \"[lindex $node 0]\""
    }
    return $attrs($attrName)
}

proc xml_get_text { l } {
    if { [llength $l] != 3 } {
        error "Expected list of 3 elements but got [llength $l]"
    }
    set child_nodes [lindex $l 2]
    foreach n $child_nodes {
        if { [llength $n] == 2 } {
            if { "[lindex $n 0]" == "#text" } {
                return [lindex $n 1]
            }
        }
    }
    return ""
}

